@@ -3,42 +3,12 @@ import soco
 import argparse
 import os
 import sys
-import ipaddress
 import pprint
-import pickle
-from soco_cli import sonos_discover
+from . import speakers
+from . import __version__
 
-
-class SpeakerList:
-    """This class handles a cache of speaker information, stored as
-    a pickle file under the user's home directory"""
-
-    def __init__(self):
-        self.config_path = os.path.expanduser("~") + "/.soco-cli"
-        if not os.path.exists(self.config_path):
-            os.mkdir(self.config_path)
-        self.pickle_file = self.config_path + "/speakers.pickle"
-        self.speakers = None
-
-    def save(self):
-        if self.speakers:
-            pickle.dump(self.speakers, open(self.pickle_file, "wb"))
-            return True
-        else:
-            return False
-
-    def load(self):
-        if os.path.exists(self.pickle_file):
-            try:
-                self.speakers = pickle.load(open(self.pickle_file, "rb"))
-            except:
-                return False
-            return True
-        else:
-            return False
-
-    def refresh(self):
-        self.speakers = sonos_discover.list_sonos_devices(socket_timeout=5)
+# Global Speakers instance
+speaker_list = None
 
 
 def error_and_exit(msg):
@@ -48,48 +18,6 @@ def error_and_exit(msg):
     os._exit(1)
 
 
-def is_ipv4_address(speaker_name):
-    try:
-        ipaddress.IPv4Network(speaker_name)
-        return True
-    except ValueError:
-        return False
-
-
-def get_speaker(
-    speaker_name,
-    use_local_speaker_list=False,
-    refresh_local_speaker_list=False,
-    require_coordinator=True,
-):
-    try:
-        if is_ipv4_address(speaker_name):
-            return soco.SoCo(speaker_name)
-        if not use_local_speaker_list:
-            # Use standard SoCo lookup
-            speaker = soco.discovery.by_name(speaker_name)
-            if speaker:
-                return speaker
-            else:
-                error_and_exit("Speaker '{}' not found.".format(speaker_name))
-        else:
-            # Use local discovery and cached speaker list if it exists
-            speaker_list = SpeakerList()
-            if not speaker_list.load() or refresh_local_speaker_list:
-                speaker_list.refresh()
-                speaker_list.save()
-            for speaker in speaker_list.speakers:
-                if speaker.speaker_name.lower() == speaker_name.lower():
-                    if require_coordinator:
-                        if speaker.is_coordinator:
-                            return soco.SoCo(speaker.ip_address)
-                    else:
-                        return soco.SoCo(speaker.ip_address)
-            error_and_exit("Speaker '{}' not found.".format(speaker_name))
-    except Exception as e:
-        error_and_exit("Exception: {}".format(str(e)))
-
-
 def print_speaker_info(speaker):
     info = speaker.get_speaker_info()
     model = info["model_name"].lower()
@@ -158,6 +86,16 @@ def pause_all(speaker):
                 pass
 
 
+def get_speaker(name, local=False):
+    if local:
+        return speaker_list.find(name)
+    else:
+        if speakers.Speakers.is_ipv4_address(name):
+            return soco.SoCo(name)
+        else:
+            return soco.discovery.by_name(name)
+
+
 def main():
     # Create the argument parser
     parser = argparse.ArgumentParser(
@@ -183,29 +121,52 @@ def main():
         help="Use the local speaker list instead of SoCo discovery",
     )
     parser.add_argument(
-        "--refresh-local-speaker_list",
+        "--refresh-local-speaker-list",
         "-r",
         action="store_true",
         default=False,
         help="Refresh the local speaker list",
     )
-
-    pp = pprint.PrettyPrinter(width=100)
+    parser.add_argument(
+        "--network-discovery-threads",
+        "-t",
+        type=int,
+        default=128,
+        help="Maximum number of threads used for Sonos network discovery",
+    )
+    parser.add_argument(
+        "--network-discovery-timeout",
+        "-n",
+        type=float,
+        default=3.0,
+        help="Network timeout for Sonos device scan (seconds)",
+    )
 
     # Parse the command line
     args = parser.parse_args()
 
+    if args.use_local_speaker_list:
+        global speaker_list
+        speaker_list = speakers.Speakers(
+            network_threads=args.network_discovery_threads,
+            network_timeout=args.network_discovery_timeout,
+        )
+        if args.refresh_local_speaker_list or not speaker_list.load():
+            speaker_list.discover()
+            speaker_list.save()
+
+    pp = pprint.PrettyPrinter(width=100)
+
     # Process the actions
     # Wrap everything in a try/except to catch all SoCo (etc.) errors
-    # ToDo: improve so there's a single action pattern and a single function to interpret it
     try:
-        speaker = get_speaker(
-            args.speaker, args.use_local_speaker_list, args.refresh_local_speaker_list
-        )
-        if not speaker:
-            error_and_exit("Speaker not found")
-        np = len(args.parameters)
         action = args.action.lower()
+        if not action in ["version"]:
+            # Some actions don't require a valid speaker
+            speaker = get_speaker(args.speaker, args.use_local_speaker_list)
+            if not speaker:
+                error_and_exit("Speaker not found")
+        np = len(args.parameters)
         # Mute ######################################################
         if action == "mute":
             if np == 0:
@@ -321,7 +282,7 @@ def main():
             else:
                 error_and_exit("Action 'line_in' takes 0, 1, or 2 parameter(s)")
         # Volume ####################################################
-        elif action in ["volume", "vol"]:
+        elif action in ["volume", "vol", "v"]:
             if np == 0:
                 print(speaker.volume)
             elif np == 1:
@@ -332,7 +293,7 @@ def main():
                     error_and_exit("Volume parameter must be from 0 to 100")
             else:
                 error_and_exit("Action 'volume' takes 0 or 1 parameter(s)")
-        elif action in ["relative_volume", "rel_vol"]:
+        elif action in ["relative_volume", "rel_vol", "relvol", "rv"]:
             if np == 1:
                 volume = int(args.parameters[0])
                 if -100 <= volume <= 100:
@@ -350,7 +311,7 @@ def main():
                     error_and_exit("Ramp parameter must be from 0 to 100")
             else:
                 error_and_exit("Action 'ramp/ramp_to_volume' requires 1 parameter")
-        elif action in ["group_volume", "group_vol"]:
+        elif action in ["group_volume", "group_vol", "gv"]:
             if np == 0:
                 print(speaker.group.volume)
             elif np == 1:
@@ -361,7 +322,7 @@ def main():
                     error_and_exit("Group Volume parameter must be from 0 to 100")
             else:
                 error_and_exit("Action 'group_volume' takes 0 or 1 parameter(s)")
-        elif action in ["group_relative_volume", "group_rel_vol"]:
+        elif action in ["group_relative_volume", "group_rel_vol", "grv"]:
             if np == 1:
                 volume = int(args.parameters[0])
                 if -100 <= volume <= 100:
@@ -513,7 +474,7 @@ def main():
                     "Action 'status_light' requires 0 or 1 parameter ('on' or 'off')"
                 )
         # Grouping ##################################################
-        elif action == "group":
+        elif action in ["group", "g"]:
             if np == 1:
                 speaker2 = get_speaker(args.parameters[0], args.use_local_speaker_list)
                 speaker.join(speaker2)
@@ -521,7 +482,7 @@ def main():
                 error_and_exit(
                     "Action 'group' requires 1 parameter (the speaker to group with"
                 )
-        elif action == "ungroup":
+        elif action in ["ungroup", "u"]:
             if np == 0:
                 speaker.unjoin()
             else:
@@ -581,6 +542,10 @@ def main():
                 speaker.separate_stereo_pair()
             else:
                 error_and_exit("Action 'unpair' requires no parameters")
+        # Version ###################################################
+        elif action == "version":
+            print("soco-cli version: {}".format(__version__))
+            print("soco version:     {}".format(soco.__version__))
         # Invalid Action ############################################
         else:
             error_and_exit("Action '{}' is not defined.".format(action))
